import math
import numpy as np
import gym
import gym.envs.classic_control.cartpole
import time
import pyglet
import __main__


class TextRender:
    def __init__(self, text):
        self.text = text

    def render(self):
        self.text.draw()


class CartPole(gym.envs.classic_control.cartpole.CartPoleEnv):
    def __init__(self):
        super().__init__()
        self.light_sensor_width = 0.100  # radius
        self.light_sensor_height = 0.030  # height above ground
        self.max_light_val = 255
        self.max_light_dist = 0.1
        self.length = 0.5  # Height of the entire robot
        self.max_power = 100  # max motor force output
        self.theta_threshold_radians = math.pi / 2
        self.encoder_ticks_per_revolution = 255
        self.wheel_radius = 0.04

        self.masscart = 4.0
        self.masspole = 1.0
        self.total_mass = (self.masspole + self.masscart)

        self.motor_power = 2  # smaller is more powerful, must be positive
        self.motor_scaling = 1
        self.static_friction = 0.1
        self.force = 0
        global label
        self.label_text = label
        self.dt = 0.2
        self.steps = 0
        self.max_steps = 800
        self.rendering = False
        self.harder = False

        self.kinematics_integrator = 'euler'

    def sensor_value(self, linear_dist):
        linear_dist = min(linear_dist, self.max_light_dist)
        sensor_reading = (self.max_light_dist - linear_dist) ** 2 / (self.max_light_dist ** 2)
        return int(sensor_reading * self.max_light_val)

    def correct_state(self, state):
        '''Converts the environment state to the self-balancing robot state'''
        pi = math.pi
        x, x_dot, theta, theta_dot = state
        left_dist = max(0, self.light_sensor_height + self.light_sensor_width * math.sin(theta + pi))
        right_dist = max(0, self.light_sensor_height + self.light_sensor_width * math.sin(-theta - pi))
        left_sensor = self.sensor_value(left_dist)
        right_sensor = self.sensor_value(right_dist)
        x /= 2 * pi
        x /= self.wheel_radius * 2
        x = int(x * self.encoder_ticks_per_revolution)
        return np.array([left_sensor, right_sensor, x])

    def reset(self):
        '''Resets the cartpole to the initial state.

        Returns a state vector consisting of
            (l, r, x)
        where `l` is the left light sensor reading, `r` is the right light
        sensor reading, and x is the current wheel encoder value.'''
        if not self.harder:
            self.state = self.np_random.uniform(low=-0.1, high=0.1, size=(4,)) * (1.0, 0.001, 1, 0.1)
        else:
            self.state = self.np_random.uniform(low=-0.1, high=0.1, size=(4,)) * (2, 0.01, 1.5, 0.1)
        self.steps_beyond_done = None
        self.steps = 0
        return self.correct_state(np.array(self.state))

    def step(self, force, ntimes=1):
        '''Performs multiple steps of simulation, by repeatedly calling _step(force) ntimes
        times'''
        dt = self.dt / ntimes
        for _ in range(ntimes - 1):
            tic = time.perf_counter()
            out = self._step(force)
            _, done, _ = out
            if done:
                return out
            # time.sleep(max(0, dt-(time.perf_counter() - tic)))
        return self._step(force)

    def _step(self, force):
        '''Does one step of simulation. Takes in an action, which can be
        move left or move right (single number, negative means left)'''
        force = max(-self.max_power, min(force, self.max_power))
        force *= self.motor_scaling
        self.force = self.force * 0.95 + force * 0.05
        state = self.state
        x, x_dot, theta, theta_dot = state
        # begin system dynamics
        sintheta = math.sin(-theta)
        costheta = math.cos(-theta)
        xacc = (1 / (self.masscart + self.masspole * (sintheta ** 2))) \
               * (force + self.masspole * sintheta * (self.length * (theta_dot ** 2) + self.gravity * costheta))

        if abs(x_dot) > self.static_friction or abs(xacc) > self.static_friction:
            thetaacc = (1 / (self.length * (self.masscart + self.masspole * (sintheta ** 2)))) \
                       * (-force * costheta - self.masspole * self.length * (theta_dot ** 2) * costheta * sintheta \
                          - (self.total_mass) * self.gravity * sintheta)

            x_dot += xacc * self.tau
            theta_dot += thetaacc * self.tau
        else:
            thetaacc = (1 / (self.length * (self.masscart + self.masspole * (sintheta ** 2)))) \
                       * (- self.masspole * self.length * (theta_dot ** 2) * costheta * sintheta \
                          - (self.total_mass) * self.gravity * sintheta)

            x_dot *= 0.8
            theta_dot += thetaacc * self.tau

        x += x_dot * self.tau
        theta += theta_dot * self.tau

        self.state = (x, x_dot, theta, theta_dot)
        self.steps += 1
        done = x < -self.x_threshold \
               or x > self.x_threshold \
               or theta < -self.theta_threshold_radians \
               or theta > self.theta_threshold_radians
        done = bool(done) or (self.steps > self.max_steps)

        if not done:
            reward = 1.0
        elif self.steps_beyond_done is None:
            # Pole just fell!
            self.steps_beyond_done = 0
            reward = 1.0
        else:
            if self.steps_beyond_done == 0:
                logger.warn(
                    "You are calling 'step()' even though this environment has already returned done = True. You should always call 'reset()' once you receive 'done = True' -- any further steps are undefined behavior.")
            self.steps_beyond_done += 1
            reward = 0.0
        if self.rendering:
            self.render()
        return self.correct_state(np.array(self.state)), done, reward

    def render(self, mode='human'):
        screen_width = 600
        screen_height = 400

        world_width = self.x_threshold * 2
        scale = screen_width / world_width
        carty = 90  # BOTTOM OF CART
        polewidth = 10.0
        polelen = scale * (2 * self.length)
        sensor_width = scale * self.light_sensor_width
        sensor_height = scale * self.light_sensor_height
        sensor_thick = 5.0
        cartwidth = 50.0
        cartheight = 30.0
        wheel_radius = self.wheel_radius * 260

        if self.viewer is None:
            from gym.envs.classic_control import rendering
            self.viewer = rendering.Viewer(screen_width, screen_height)
            l, r, t, b = -cartwidth / 2, cartwidth / 2, cartheight / 2, -cartheight / 2
            axleoffset = 0  # cartheight/4.0
            #            cart = rendering.FilledPolygon([(l,b), (l,t), (r,t), (r,b)])
            cart = rendering.make_circle(radius=wheel_radius, res=30, filled=True)
            self.carttrans = rendering.Transform()
            cart.add_attr(self.carttrans)
            self.viewer.add_geom(cart)
            l, r, t, b = -polewidth / 2, polewidth / 2, polelen - polewidth / 2, -polewidth / 2
            pole = rendering.FilledPolygon([(l, b), (l, t), (r, t), (r, b)])
            pole.set_color(.8, .6, .4)
            self.poletrans = rendering.Transform(translation=(0, axleoffset))
            pole.add_attr(self.poletrans)
            pole.add_attr(self.carttrans)
            self.viewer.add_geom(pole)
            self.axle = rendering.make_circle(polewidth / 2)
            self.axle.add_attr(self.poletrans)
            self.axle.add_attr(self.carttrans)
            self.axle.set_color(.5, .5, .8)
            self.viewer.add_geom(self.axle)

            ls, rs, ts, bs = -sensor_width, sensor_width, sensor_thick / 2, -sensor_thick / 2
            sensors = rendering.FilledPolygon([(ls, bs), (ls, ts), (rs, ts), (rs, bs)])
            sensors.set_color(.6, .3, .2)
            self.sensortrans = rendering.Transform(translation=(0,
                                                                axleoffset + sensor_height))
            sensors.add_attr(self.poletrans)
            sensors.add_attr(self.carttrans)
            self.viewer.add_geom(sensors)
            self.track = rendering.Line((0, carty), (screen_width, carty))
            self.track.set_color(0, 0, 0)
            self.viewer.add_geom(self.track)

            self._pole_geom = pole
            self.label = TextRender(pyglet.text.Label(self.label_text,
                                                      font_size=12,
                                                      x=screen_width / 2, y=20,
                                                      anchor_x='center', anchor_y='center',
                                                      color=(0, 0, 0, 150)
                                                      ))
            self.viewer.add_geom(self.label)

        if self.state is None: return None

        # Edit the pole polygon vertex
        pole = self._pole_geom
        l, r, t, b = -polewidth / 2, polewidth / 2, polelen - polewidth / 2, -polewidth / 2
        pole.v = [(l, b), (l, t), (r, t), (r, b)]

        x = self.state
        cartx = x[0] * scale + screen_width / 2.0  # MIDDLE OF CART
        self.carttrans.set_translation(cartx, carty + wheel_radius)
        self.poletrans.set_rotation(-x[2])

        return self.viewer.render(return_rgb_array=mode == 'rgb_array')


try:
    label = __main__.__file__
except AttributeError:
    label = 'interactive terminal'

if __name__ == '__main__':
    import random

    cartpole = CartPole()
    cartpole.rendering = True
    for _ in range(5):
        done = False
        state = cartpole.reset()
        while not done:
            tic = time.perf_counter()
            action = (random.random() - 0.5) * 100.0
            state, done, _ = cartpole.step(action)
    cartpole.close()
